/*
 * ----------------------------------------------------------------------------
 *
 * zfconn.c -- zzzforward ZFConn connection data type
 * Copyright (C) 2004 Fabian "zzznowman" Pietsch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ----------------------------------------------------------------------------
 *
 */

/* needed for e.g. union of all known struct sockaddr_* */
#define _GNU_SOURCE

#include <errno.h>	/* errno				*/
#include <err.h>	/* warn()				*/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>	/* select()				*/
#include <sys/types.h>	/* select()				*/
#include <unistd.h>	/* read(), write(), usleep(), select()	*/
#include <fcntl.h>	/* fcntl() for non-blocking I/O		*/

#include <netdb.h>	/* gethostbyname()			*/
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* inet_aton()				*/


#include "zzzforward.h"
#include "memory.h"
#include "zfconn.h"


/* linked list to store every ZFConn */
ZFConn *connections;


/*
 * connection-related helper functions
 */

/* connect to desthost:destport */
int zf_connect(int *s, SysSA *sa) {
	struct hostent *he;

	/* lookup destination host */
	if ((he = gethostbyname(desthost)) == NULL) {
		fprintf(stderr, "error: gethostbyname() failed (%s)\n",
		  (h_errno == HOST_NOT_FOUND) ? "host not found"
		  : "unknown error");
		return -1;
	}

	/* create a socket */
	if ((*s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("error: socket()");
		return -1;
	}

	/* prepare sa */
	sa->sin_family = AF_INET;
	sa->sin_port = htons(atol(destport));
	sa->sin_addr = *((struct in_addr *)he->h_addr_list[0]);

	/* connect the socket */
	if (connect(*s, sa, sizeof(*sa))) {
		perror("error: connect()");
		close(*s);
		return -1;
	}

	/* set s nonblocking */
	if (fcntl(*s, F_SETFL, fcntl(*s, F_GETFL) | O_NONBLOCK)) {
		perror("error: fcntl()");
		close(*s);
		return -1;
	}

	/* success */
	return 0;
}

/* perform steps necessary for accepting incoming connections */
int zf_listen(int *s, int port) {
	SysSA sa;

	/* create a socket */
	if ((*s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("error: socket()");
		return -1;
	}

	/* set s nonblocking
	 * this will not be inherited by accept()ed sockets,
	 * but accept() won't block
	 */
	if (fcntl(*s, F_SETFL, fcntl(*s, F_GETFL) | O_NONBLOCK)) {
		perror("error: fcntl()");
		close(*s);
		return -1;
	}

	/* prepare sa */
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = INADDR_ANY;

	/* bind on specified port */
	if (bind(*s, &sa, sizeof(sa))) {
		perror("error: bind()");
		close(*s);
		return -1;
	}

	/* finally perform listen() call on the socket */
	if (listen(*s, 3)) {
		perror("error: listen()");
		close(*s);
		return -1;
	}

	/* success */
	return 0;
}

/* zf_select() -- sleep until some socket state changes */
int zf_select(int listens) {
	fd_set set;
	ZFConn *conn = connections;

	/* initialize set of fds monitored for reading */
	FD_ZERO(&set);
	FD_SET(listens, &set);
	while (conn) {
		FD_SET(conn->client_s, &set);
		FD_SET(conn->server_s, &set);
		conn = conn->next;
	}

	/* wait for activity on any socket */
	if (select(FD_SETSIZE, &set, NULL, NULL, NULL) == -1) {
		perror("error: select()");
		return -1;
	}

	/* new incoming connection? */
	if (FD_ISSET(listens, &set))
		return 1;

	/* success */
	return 0;
}


/*
 * data structure handling functions
 */

/* zfc_new() -- create new ZFConn from listening socket */
ZFConn *zfc_new(int listens) {
	static int nextid;
	int s;
	SysSA sa;
	size_t sa_len = sizeof(sa);
	ZFConn *conn;

	/* try accept()ing a connection */
	if ((s = accept(listens, &sa, &sa_len)) == -1) {
		/* no incoming connections on non-blocking I/O? */
		if (errno != EAGAIN)
			/* a real error occured */
			perror("error: zfc_new(): accept()");

		/* fail (no new connection) */
		return NULL;
	}
	if (sa_len != sizeof(sa)) {
		fprintf(stderr, "error: zfc_new(): accept() returned unknown client data structure of size %d\n", sa_len);
		close(s);
		return NULL;
	}

	/* set new connection non-blocking */
	if (fcntl(s, F_SETFL, fcntl(s, F_GETFL) | O_NONBLOCK)) {
		perror("error: zfc_new(): fcntl()");
		close(s);
		return NULL;
	}

	/* allocate new connection information data structure */
	if ((conn = NEW(ZFConn)) == NULL) {
		perror("error: zfc_new(): malloc()");
		close(s);
		return NULL;
	}

	/* fill in connection information */
	conn->id = nextid++;
	conn->status= ZFCS_CCONNECTED;
	conn->client_s = s;
	conn->client_a = sa;
	conn->next = connections;

	/* insert connection information into list of active connections */
	connections = conn;

	/* be verbose about what happened */
	printf("new connection %d from %s:%d\n", conn->id,
	  inet_ntoa(conn->client_a.sin_addr),
	  ntohs(conn->client_a.sin_port));

	/* connect to real server */
	if (zf_connect(&conn->server_s, &conn->server_a)) {
		fprintf(stderr, "error: [%d] connecting to real server failed\n", conn->id);
#if 0
		if (close(conn->client_s))
			warn("error: [%d] close() on client socket", conn->id);
		else
			/* clear "client connected" ZFConn status bit */
			conn->status = conn->status & ~ZFCS_CCONNECTED;
		close(s);
#endif
		return conn;
	}
	conn->status = conn->status | ZFCS_SCONNECTED;

	/* be verbose about what happened */
	printf("[%d] connected to server %s:%d\n", conn->id,
	  inet_ntoa(conn->server_a.sin_addr),
	  ntohs(conn->server_a.sin_port));

	/* success */
	return conn;
}

/* zfc_free() -- free a ZFConn */
void zfc_free(ZFConn *conn) {
	ZFConn *parent;

	assert(conn != NULL);

	/* disconnect if still connected */
	if (conn->status & ZFCS_CONNECTED)
		zfc_disconnect(conn);

	/* remove conn from list of connections */
	if ((parent = zfc_parent(conn)))
		parent->next = conn->next;
	else {
		assert(connections == conn);
		connections = conn->next;
	}

	/* free this ZFConn */
	FREE(conn);
}

/* disconnect an active connection */
int zfc_disconnect(ZFConn *conn) {
	assert(conn != NULL);

	/* check if really still connected */
	if (!(conn->status & ZFCS_CONNECTED)) {
		fprintf(stderr, "warning: [%d] zfc_disconnect(): not connected anymore\n", conn->id);
		return 0;
	}

	/* be verbose about what's happening */
	printf("[%d] disconnecting\n", conn->id);

	/* close connection to real server? */
	if (conn->status & ZFCS_SCONNECTED) {
		if (close(conn->server_s))
			warn("error: [%d] close() on server socket", conn->id);
		else
			/* clear "server connected" ZFConn status bit */
			conn->status = conn->status & ~ZFCS_SCONNECTED;
	}

	/* close connection to client? */
	if (conn->status & ZFCS_CCONNECTED) {
		if (close(conn->client_s))
			warn("error: [%d] close() on client socket", conn->id);
		else
			/* clear "client connected" ZFConn status bit */
			conn->status = conn->status & ~ZFCS_CCONNECTED;
	}

	/* success if not connected anymore */
	return (conn->status & ZFCS_CONNECTED);
}

/* __zfc_forward_oneway() -- forward data from one socket to another */
static int __zfc_forward_oneway(int src, int dst) {
	char buffer[65535], *bufp;
	int incount, outcount, tries;

	switch (incount = read(src, buffer, sizeof(buffer))) {
	case -1: /* error */
		/* no data available on non-blocking IO? */
		if (errno != EAGAIN)
			/* other error => fail */
			return -1;

		/* skip this transfer direction (lack of data) */
		break;
	case  0: /* end of file */
		/* fail */
		return -1;
	default: /* received some data */
		/* try sending data until everything is sent */
		bufp = buffer; tries = 0;
		while ((outcount = write(dst, bufp, incount)) < incount) {
			switch (outcount) {
			case -1: /* error */
				/* no buffer space available on non-blocking IO? */
				if (errno != EAGAIN)
					/* other error => fail */
					return -1;

				/* TRIES_MAX exceeded? */
				if (++tries > TRIES_MAX)
					/* timeout */
					return -2;

				/* delay for TRIES_DELAY milliseconds,
				 * then try again until some "real" error occurs
				 */
				usleep(TRIES_DELAY);
				continue;
			case  0: /* "nothing was written" => ? */
			default: /* sent some data */
				bufp += outcount;
				incount -= outcount;
			}
		}
	}

	/* success */
	return 0;
}

/* zfc_forward() -- forward data between ZFConn's client & server */
int zfc_forward(ZFConn *conn) {
	int *fdinp  = &(conn->server_s),
	    *fdoutp = &(conn->client_s);

	assert(conn != NULL);

	/* check if this ZFConn really is fully connected */
	if ((conn->status & ZFCS_CONNECTED) != ZFCS_CONNECTED) {
		fprintf(stderr, "warning: [%d] zfc_forward(): not fully connected\n", conn->id);
		return -1;  /* failure */
	}

__ONEWAY__:
	switch (__zfc_forward_oneway(*fdinp, *fdoutp)) {
	case  0:  /* success */
		break;
	case -2:  /* timeout */
		fprintf(stderr, "error: [%d] %s timeout\n", conn->id,
		  (*fdoutp == conn->server_s) ? "server" : "client");
	default:  /* general error */
		/* failure */
		return -1;
	}

	/* first run? */
	if (*fdinp == conn->server_s) {
		/* start over again with source and destination exchanged */
		fdinp  = &conn->client_s;
		fdoutp = &conn->server_s;
		goto __ONEWAY__;
	}

	/* success */
	return 0;
}

/* zfc_parent() -- find ZFConn's parent */
ZFConn *zfc_parent(ZFConn *conn) {
	ZFConn *cur = connections;  /* current ZFConn during enumeration */

	assert(conn != NULL);

	/* enumerate through list of connections */
	while (cur) {
		if (cur->next == conn)
			/* parent found */
			return cur;
		cur = cur->next;
	}

	/* no parent found */
	if (connections != conn)
		fprintf(stderr, "zfc_parent(): warning: conn is not in global list of connections\n");

	return NULL;
}
