/*
 * ----------------------------------------------------------------------------
 *
 * zzzforward.c -- listen on a socket and forward incoming connections
 * Copyright (C) 2003, 2004 Fabian "zzznowman" Pietsch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ----------------------------------------------------------------------------
 *
 */

#include <errno.h>       /* errno                */
#include <unistd.h>      /* getopt()             */
#include <stdio.h>       /* printf() and friends */
#include <stdlib.h>      /* exit(), strtol()     */
#include <netinet/in.h>  /* needed for zfconn.h  */


#include "zzzforward.h"
#include "zfconn.h"


/*
 * global variables
 */

char *desthost;                  /* destination to forward clients to    */
char *destport;

int listenport = ZF_LISTENPORT;  /* port zzzforward will be listening on */
int listens;                     /* listening socket for listenport      */


/* zf_forward() -- forward pending data between all active connections */
int zf_forward() {
	ZFConn *conn = connections;

	/* walk through all active connections */
	while (conn) {
		/* save next element in case this one will be freed */
		ZFConn *conn_next = conn->next;

		/* perform a single transfer cycle */
		if (zfc_forward(conn)) {
			/* error => close connection */
			fprintf(stderr, "[%d] transmission error -- dropping connection\n", conn->id);
			zfc_free(conn);
		}

		/* step ahead in linked list */
		conn = conn_next;
	}

	/* success */
	return 0;
}

/* zf_sleep() -- sleep and perform necessary actions after awaking */
int zf_sleep() {
	ZFConn *conn;

	/* sleep until some socket state changes */
	switch (zf_select(listens)) {
	case -1:  /* error */
		fprintf(stderr, "error: zfc_select() failed -- exiting\n");
		return -1;
	case  0:  /* new input, but no new incoming connection */
		break;
	case  1:  /* new incoming connection */
		/* accept new connection */
		if ((conn = zfc_new(listens)) == NULL) {
			fprintf(stderr, "error: zfc_new() failed -- exiting\n");
			return -1;
		}

		/* are both parts of the connection fully established? */
		if ((conn->status & ZFCS_CONNECTED) != ZFCS_CONNECTED) {
			/* drop connection */
			fprintf(stderr, "[%d] connection error -- dropping connection\n", conn->id);
			zfc_free(conn);
		}
		break;
	}

	/* success */
	return 0;
}

/* zf_usage() -- print usage information, then exit program */
void zf_usage(FILE *stream, int exit_status) {
	/* fputs() doesn't add a trailing '\n' */
	fputs("\
Usage: zzzforward [-h] [-l LISTENPORT] DESTHOST DESTPORT\n\
Options:\n\
  -h             this usage information\n\
  -l LISTENPORT  listen on LISTENPORT (must be numeric currently,\n\
                 no lookups are performed)\n\
  -V             output version information\n", stream);
	exit(exit_status);
}

/* zf_version() -- print out version information, then exit program */
void zf_version() {
	/* puts() adds a trailing '\n' automatically */
	puts(PRODUCT_VERSION);
	exit(EXIT_SUCCESS);
}

/* program entry point */
int main(int argc, char **argv) {
	char opt, *tail;

	/* set stdout line-buffered */
	setvbuf(stdout, NULL, _IOLBF, 0);

	/* process command line options */
	while ((opt = getopt(argc, argv, "hl:V")) != -1) {
		switch (opt) {
		case 'h':  /* help */
			zf_usage(stdout, EXIT_SUCCESS);
			break;
		case 'l':  /* listen port */
			/* convert listen port from string to long int */
			errno = 0;
			listenport = strtol(optarg, &tail, 0);
			if (errno || (*tail != '\0')) {
				fprintf(stderr, "error: invalid listen port: %s\n", optarg);
				zf_usage(stderr, EXIT_FAILURE);
			}
			break;
		case 'V':
			zf_version();
			break;
		case '?':  /* bad option or missing argument */
			zf_usage(stderr, EXIT_FAILURE);
			break;
		default:  /* error */
			abort();
			break;
		}
	}

	/* use trailing 2 arguments as destination specification */
	if (argc - optind != 2)
		zf_usage(stderr, EXIT_FAILURE);
	desthost = argv[optind];
	destport = argv[optind + 1];

	/* listen on listenport */
	if (zf_listen(&listens, listenport))
		exit(EXIT_FAILURE);

	/* be verbose about our existence */
	printf(PRODUCT_VERSION " listening on port %d\n", listenport);

	/* enter main loop */
	while (1) {
		if (zf_forward())
			exit(EXIT_FAILURE);
		if (zf_sleep())
			exit(EXIT_FAILURE);
	}

	/* success (well, currently won't ever be reached.. *hide*) */
	exit(EXIT_SUCCESS);
}
